' 23/07/2017
' Thermistor linearisation using a Stein-Hart equasion
'
' HOW TO USE the Declare Variables Library.
'
' The program below is based on a 10K Ohm resistor as part of the thermistor's circuit: -
'
'  +5 Volts
' --------
'    |
'    \
'    /  10k Ohm 1% (or better) Resistor
'    \
'    /
'    |
'    o-------------O To PORTA.0 of the PICmicro
'    |
'    \
'   o/
'    \  Thermistor
'    /
'    |
'    |
'  -----
'   ---
'    -
'-------------------------------------------------------------------------------
    Include "Proton18_20.Inc"
    Declare Optimiser_Level = 3
    Declare Dead_Code_Remove = On

    Declare Float_Display_Type = Fast                        ' Use the larger but faster floating point display routine

    Include "D:\MyIncFiles\DeclareLocalVars.inc"
'    Include "INC\DeclareLocalVars.Inc"
'-------------------------------------------------------------------------------
'    Dim MedianSwap As Word                            ' Temporary Median Filter variable for swapping
'    Dim MedianIndex As Byte                           ' Holds the position in the Bubble sort
'    Dim MedianSwapOccured As Bit                      ' Indicates if the Bubble sort is complete
    Dim ThermistorValue As Word                       ' [Global Variable] Holds the raw thermistor ADC reading
'    Dim Resistance As Float                           ' Holds the resistance of the thermistor for a given Temperature
    Dim Temperature As Float                          ' [Global Variable] Holds the floating point Temperature in degrees Centigrade
'    Dim TempFloat As Float                            ' Temporary variable for the thermistor linearisation
'    Dim LN As Float                                   ' Used to hold the LOG of the thermistor value
'
'-------------------------------------------------------------------------------
' The MedianSample array declaration and its associated constants (below) should be placed at the end of the Dim list
'
    Symbol MedianSamplesToTake = 21                   ' The amount of samples to take
    #if (MedianSamplesToTake > 127)
         #error "Constant MedianSamplesToTake cannot be greater than 127"
    #endif
    Symbol MiddleSample = (MedianSamplesToTake / 2) + 1 ' The middle sample of the samples taken
    Dim MedianSample[MedianSamplesToTake] As Word     ' Create an array to hold the range samples

    Dim Ad_Result As ADRESL.Word                      ' Alias the 16-bit ADRESL\H register as Ad_Result
'
' Constant values used for the Stein-Hart thermistor linearisation routine
'
    Symbol Quanta = 5000.0 / 1024.0                   ' Quatasising constant for ADC reading to Voltage calculation
    Symbol T25 = 0.0033540                            ' Thermistor's t25 value
    Symbol B = 1.0 / 3489.0                           ' Thermistor's B value
'
' The Stein-Hart linearising equation below is only suitable for describing a restricted range around
' the rated Temperature or resistance with sufficient accuracy.
'
'                 1
' Tt = ------------------------
'      1/B * Ln(Rt/R25) + 1/T25
'
' The B values for common NTC thermistors range from 2000K through to 5000K.
' For this application, the popular, and more complicated, Stein-Hart equation is more accurate.
'
' The thermistor used did not provide the B value within its datasheet, but did give the T25 value.
' i.e. resistance at 25 degrees Centigrade.
' To calculate the B value use the equation below.
'
'      Tt * T25        R25
' B = ---------- * Ln ------
'      Tt - T25        R100
'
' or: -
'
'      (25 + 273.13) * (100 + 273.15)        R25
' B = -------------------------------- * Ln ------
'                    75                      R100
'
' giving: -
'
'                  R25
' B = 1483.4 * Ln ------
'                  R100
'
' Thus, knowing the resistance at 25 degrees Centigrade and also at 100 degrees Centigrade
' we can use the above calculation to find the value of B for a given thermistor.
'===============================================================================
    GoTo MainProgram    ' Jump over the subroutines

'-------------------------------------------------------------------------------
' Subroutines Start Here
'-------------------------------------------------------------------------------
' Create a Median filter to make an educated guess as to what is the true reading from the Temperature readings taken.
' The routine below is a Bubble Sort, that arranges all the samples in ascending order within the array MedianSampleS.
' The middle sample is then extracted. This should eliminate spurious readings from the edges.
'
' Input     : Array MedianSample holds "MedianSamplesToTake" Temperature readings taken
' Output    : ThermistorValue = the median (middle value) of the samples taken
'
' Local System variables for this Macro.
    LocalBit0(MedianSwapOccured)    ' Indicates if the Bubble sort is complete
    LocalByte0(MedianIndex)         ' Holds the position in the Bubble sort
    LocalWord0(MedianSwap)          ' Temporary Median Filter variable for swapping

$define MedianFilter() GoSub MedianFilter_Sub

MedianFilter_Sub:
    Repeat
        MedianSwapOccured = 0                                                 ' Clear flag that indicates swap.
        MedianIndex = 0
        Repeat                                                                ' For each cell of the array...
            If MedianSample[MedianIndex] > MedianSample[MedianIndex + 1] Then ' Move larger values up.
                MedianSwap = MedianSample[MedianIndex]                        ' ..by swapping them.
                MedianSample[MedianIndex] = MedianSample[MedianIndex + 1]
                MedianSample[MedianIndex + 1] = MedianSwap
                MedianSwapOccured = 1                                         ' Set bit if swap occurred.
            EndIf
            Inc MedianIndex
        Until MedianIndex = MedianSamplesToTake                               ' Check out next cell of the array.
    Until MedianSwapOccured = 0                                               ' Keep sorting until no more swaps.
    ThermistorValue = MedianSample[MiddleSample]                             ' Extract the middle sample's value
    ' Because the "ThermistorValue" is used in other routine then it is a Global Variable.
    Return

'-------------------------------------------------------------------------------
' Read the thermistor Temperature sensor on AN0 (PortA.0) and linearise the result
' Input      : None
' Output     : Temperature = Temperature in degrees Centigrade
' Notes      : The thermistor is sampled by the ADC with 10-bit resolution several times
'            : based on the constant "MedianSamplesToTake".
'            : A Median filter is then called to filter out any noise
'            : Then the raw (filtered) thermistor value is run through a Stein-Hart calculation
'            : which linearises the value into degrees Centigrade.
'            : Accuracy depends on the thermistor and its accompanying 10K Ohm resistor.
'
' Local System variables for this Macro.
    LocalFloat0(Resistance)     ' Holds the resistance of the thermistor for a given Temperature
    LocalFloat1(TempFloat)      ' Temporary variable for the thermistor linearisation
    LocalFloat2(LN)             ' Used to hold the LOG of the thermistor value

$define ReadTempSensor() GoSub ReadTempSensor_Sub

ReadTempSensor_Sub:
    ADCON1 = %10000000                              ' Make PORTA and PORTE all analogue and Right justify the ADC result
    ADCON0 = %10000001                              ' Set ADC: FRC/32 OSC, AN0, GO_DONE = 0, ADC Powered up
    For MedianIndex = 0 To MedianSamplesToTake - 1  ' Form a loop for the amount of Temperature readings to take
        DelayUS 5                                   ' Wait for the ADC's internal capacitors to charge
        ADCON0bits_GO_DONE = 1                      ' \ Take an ADC reading
        Repeat : Until ADCON0bits_GO_DONE = 0       ' / Wait for the ADC conversion to finish
        MedianSample[MedianIndex] = Ad_Result       ' Add the reading to the median array
    Next                                            ' Close the loop
    MedianFilter()                                  ' Perform a median filter on the temperatures read
'
' Perform a Stein-Hart calculation to linearise the thermistor into degrees Centigrade
'
    Temperature = ThermistorValue * Quanta          ' \
    TempFloat = 5000 - Temperature                  '   Find the resistance across the thermistor
    Resistance = (Temperature * 10000) / TempFloat  ' /
    LN = Log (Resistance / 10000)                   ' \
    Temperature = (B * LN) + T25                    ' / Linearise the result using the stein-hart algorithm
    Temperature = (1 / Temperature) - 273.15        ' Convert from Kelvin to Centigrade
    ' Because the "Temperature" is used in the MAIN then it is a Global Variable.
    ADCON1 = 7                                      ' Make PORTA and PORTE all digital
    Return                                          ' Then return from the subroutine

'===============================================================================
' The Main Demo Program Loop Starts Here
MainProgram:
    SMLocalByte0(Aloop)                             ' Example Using a Local Variable in main code.

    DelayMS 100                                     ' Wait for the LCD to stabilise
    Cls                                             ' Clear the LCD
    Print 254,72,%00000110,%00001001,%00001001,%00000110,0,0,0,0 ' Build the LCD degrees symbol in character 1

    For Aloop = 0 To 100                            ' Create a loop
        ReadTempSensor()                            ' Read and linearise the thermistor
        Print At 1,1,Dec1 Temperature,1,"C  "       ' Display the result on the LCD
        DelayMS 100                                 ' Wait a while
    Next                                            ' Close the loop
    GoTo MainProgram
'-------------------------------------------------------------------------------
    End


